#include "MarkerBasedAR.h"

void performThreshold(const cv::Mat& grayscale, cv::Mat& thresholdImg)
{
	cv::adaptiveThreshold(
		grayscale,
		thresholdImg,
		255,
		cv::ADAPTIVE_THRESH_GAUSSIAN_C,
		cv::THRESH_BINARY_INV,
		7,
		7
		);
}

void findContours(const cv::Mat& thresholdImg,
	std::vector< std::vector<cv::Point> >& contours,
	int minContourPointsAllowed)
{
	std::vector< std::vector<cv::Point> > allContours;
	cv::findContours(thresholdImg, allContours, CV_RETR_LIST, CV_CHAIN_APPROX_NONE);
	contours.clear();
	for (size_t i = 0; i<allContours.size(); i++)
	{
		int contourSize = allContours[i].size();
		if (contourSize > minContourPointsAllowed)
		{
			contours.push_back(allContours[i]);
		}
	}
}


void findMarkerQuads(std::vector< std::vector<cv::Point> >& contours,
	std::vector<Marker>& markers,
	float minContourLengthAllowed)
{
	std::vector<cv::Point> approxCurve;
	
	for (size_t i = 0; i<contours.size(); i++)
	{
		// Approximate to a polygon
		double eps = contours[i].size() * 0.05;
		cv::approxPolyDP(contours[i], approxCurve, eps, true);
		// We interested only in polygons that contains only four points
		if (approxCurve.size() != 4)
			continue;
		// And they have to be convex
		if (!cv::isContourConvex(approxCurve))
			continue;
		// Ensure that the distance between consecutive points is large enough
		float minDist = 100000000000.0f;
		for (int i = 0; i < 4; i++)
		{
			cv::Point side = approxCurve[i] - approxCurve[(i + 1) % 4];
			float squaredSideLength = side.dot(side);
			minDist = std::min(minDist, squaredSideLength);
		}
		// Check that distance is not very small
		if (minDist < minContourLengthAllowed)
			continue;
		// All tests are passed. Save marker candidate:
		Marker m;
		for (int i = 0; i<4; i++)
			m.points.push_back(cv::Point2f(approxCurve[i].x, approxCurve[i].y));
		// Sort the points in anti-clockwise order
		// Trace a line between the first and second point.
		// If the third point is at the right side, then the points	are anticlockwise
		cv::Point v1 = m.points[1] - m.points[0];
		cv::Point v2 = m.points[2] - m.points[0];
		double o = (v1.x * v2.y) - (v1.y * v2.x);
		if (o < 0.0) //if the third point is in the left side, then sort in anti - clockwise order
			std::swap(m.points[1], m.points[3]);
		markers.push_back(m);
	}
}

float perimeter(const Marker& m)
{
	float perim = 0;
	for (int i = 0; i < m.points.size()-1; ++i)
	{
		cv::Point p1 = m.points[i];
		cv::Point p2 = m.points[i + 1];
		cv::Point edge = p2 - p1;
		perim += sqrt(edge.dot(edge));
	}
	return perim;
}

void removeMarkerDuplicates(std::vector<Marker>& markersIn,
	std::vector<Marker>& markersOut)
{
	// Remove these elements which corners are too close to each other.
	// First detect candidates for removal:
	std::vector< std::pair<int, int> > tooNearCandidates;
	for (size_t i = 0; i< markersIn.size(); i++)
	{
		const Marker& m1 = markersIn[i];
		//calculate the average distance of each corner to the nearest corner of the other marker candidate
		for (size_t j = i + 1; j< markersIn.size(); j++)
		{
			const Marker& m2 = markersIn[j];
			float distSquared = 0;
			for (int c = 0; c < 4; c++)
			{
				cv::Point v = m1.points[c] - m2.points[c];
				distSquared += v.dot(v);
			}
			distSquared /= 4;
			if (distSquared < 100)
			{
				tooNearCandidates.push_back(std::pair<int, int>(i, j));
			}
			}
	}
	// Mark for removal the element of the pair with smaller perimeter
	std::vector<bool> removalMask(markersIn.size(), false);
	for (size_t i = 0; i<tooNearCandidates.size(); i++)
	{
		float p1 = perimeter(markersIn[tooNearCandidates[i].first]);
		float p2 = perimeter(markersIn[tooNearCandidates[i].second]);
		size_t removalIndex;
		if (p1 > p2)
			removalIndex = tooNearCandidates[i].second;
		else
			removalIndex = tooNearCandidates[i].first;
		removalMask[removalIndex] = true;
	}
	// Return candidates
	markersOut.clear();
	for (size_t i = 0; i< markersIn.size(); i++)
	{
		if (!removalMask[i])
			markersOut.push_back(markersIn[i]);
	}
}


void removePerspective(const cv::Mat &image, std::vector<Marker>& markers, std::vector<cv::Mat>& canonicalMarkers)
{
	cv::Size markerSize(100, 100);
	std::vector<cv::Point2f> origPoints;
	origPoints.push_back(cv::Point2f(0, 0));
	origPoints.push_back(cv::Point2f(markerSize.width - 1, 0));
	origPoints.push_back(cv::Point2f(markerSize.width - 1, markerSize.height - 1));
	origPoints.push_back(cv::Point2f(0, markerSize.height - 1));

	cv::Mat canonicalMarker;

	for (size_t iMarker = 0; iMarker < markers.size(); ++iMarker)
	{
		Marker& marker = markers[iMarker];
		// getPerspectiveTransform requires points with floating point coordinates
		std::vector<cv::Point2f> points;
		for (int i = 0; i < marker.points.size(); ++i)
		{
			points.push_back(marker.points[i]);
		}

		// Find the perspective transfomation that brings current marker to rectangular form
		markers[iMarker].perspM = cv::getPerspectiveTransform(points, origPoints);
		cv::invert(markers[iMarker].perspM, markers[iMarker].perspMInv);
		// Transform image to get a canonical marker image
		cv::warpPerspective(image, canonicalMarker, markers[iMarker].perspM, markerSize);

		canonicalMarkers.push_back(canonicalMarker);
	}
}

int getHammingDistance(cv::Mat& bits)
{
	int ids[4][5] =
	{
		{ 1, 0, 0, 0, 0 },
		{ 1, 0, 1, 1, 1 },
		{ 0, 1, 0, 0, 1 },
		{ 0, 1, 1, 1, 0 }
	};

	int dist = 0;

	for (int y = 0; y<5; y++)
	{
		int minSum = 1e5; //hamming distance to each possible word

		for (int p = 0; p<4; p++)
		{
			int sum = 0;
			for (int x = 0; x<5; x++)
			{
				sum += bits.at<uchar>(y, x) == ids[p][x] ? 0 : 1;
			}

			if (minSum>sum)
				minSum = sum;
		}
		dist += minSum;
	}
	return dist;
}

int mat2id(const cv::Mat &bits)
{
	int val = 0;
	for (int y = 0; y<5; y++)
	{
		val <<= 1;
		if (bits.at<uchar>(y, 1)) val |= 1;
		val <<= 1;
		if (bits.at<uchar>(y, 3)) val |= 1;
	}
	return val;
}

cv::Mat rotate(cv::Mat in)
{
	cv::Mat out;
	in.copyTo(out);
	for (int i = 0; i<in.rows; i++)
	{
		for (int j = 0; j<in.cols; j++)
		{
			out.at<uchar>(i, j) = in.at<uchar>(in.cols - j - 1, i);
		}
	}
	return out;
}


void getMarkerIds(std::vector<cv::Mat>& canonicalMarkers, std::vector<Marker>& markers)
{
	for (int i = 0; i < canonicalMarkers.size(); ++i)
	{
		int cellSize = canonicalMarkers[i].rows / 7;

		cv::Mat bitMatrix = cv::Mat::zeros(5, 5, CV_8UC1);

		for (int y = 0; y < 5; y++)
		{
			for (int x = 0; x<5; x++)
			{
				int cellX = (x + 1)*cellSize;
				int cellY = (y + 1)*cellSize;
				cv::Mat cell = canonicalMarkers[i](cv::Rect(cellX, cellY, cellSize, cellSize));
				int nZ = cv::countNonZero(cell);
				if (nZ>(cellSize*cellSize) / 2)
					bitMatrix.at<uchar>(y, x) = 1;
			}
		}

		cv::Mat rotations[4];
		int distances[4];
		rotations[0] = bitMatrix;
		distances[0] = getHammingDistance(rotations[0]);
		std::pair<int, int> minDist(distances[0], 0);
		for (int i = 1; i < 4; i++)
		{
			//get the hamming distance to the nearest possible word
			rotations[i] = rotate(rotations[i - 1]);
			distances[i] = getHammingDistance(rotations[i]);
			if (distances[i] < minDist.first)
			{
				minDist.first = distances[i];
				minDist.second = i;
			}
		}

		if (minDist.first == 0)
		{
			int markerId = mat2id(rotations[minDist.second]);
			markers[i].id = markerId;
		}
		else
			markers[i].id =  -1;
	}
}

CameraCalibration::CameraCalibration()
{
	float matdata[] = { 653.1738477554757,                 0, 327.5578948350118,
						  0,               652.6302381356953, 190.7371570195519,
						  0,                               0,   1 };
	//ideal parameters?
	//float matdata[] = { 650, 0, 320,
	//	0, 650, 180,
	//	0, 0, 1 };

	m_intrinsic = cv::Mat1f(3, 3,matdata);

	//m_intrinsic = m_intrinsic.t();
	m_distorsion.resize(5);

	m_distorsion[0] =  -0.2285804148616529;
	m_distorsion[1] = 3.912081122786242;
	m_distorsion[2] = 0.005855814037156479;
	m_distorsion[3] = -0.000151208767404744;
	m_distorsion[4] = -17.89318815928614;

	//ideal parameters?
	//m_distorsion[0] = 0;
	//m_distorsion[1] = 1;
	//m_distorsion[2] = 0;
	//m_distorsion[3] = 0;
	//m_distorsion[4] = 1;
}

void estimatePosition(std::vector<Marker>& detectedMarkers, CameraCalibration& calibration)
{
	std::vector<cv::Point3f> m_markerCorners3d;
	m_markerCorners3d.push_back(cv::Point3f(-0.5,0,-0.5));
	m_markerCorners3d.push_back(cv::Point3f( 0.5,0,-0.5));
	m_markerCorners3d.push_back(cv::Point3f( 0.5,0, 0.5));
	m_markerCorners3d.push_back(cv::Point3f(-0.5,0, 0.5));
	for (size_t i = 0; i < detectedMarkers.size(); i++)
	{
		Marker& m = detectedMarkers[i];
		cv::Mat Rvec;
		cv::Mat_<float> Tvec;
		cv::Mat raux, taux;
		std::vector<cv::Point2f> markerPoints;
		for (int mp = 0; mp < m.points.size(); ++mp)
		{
			markerPoints.push_back(m.points[mp]);
		}
		cv::solvePnP(m_markerCorners3d, markerPoints, calibration.m_intrinsic, cv::InputArray(calibration.m_distorsion), raux, taux);
		raux.convertTo(Rvec, CV_32F);
		taux.convertTo(Tvec, CV_32F);
		cv::Mat_<float> rotMat(3, 3);
		cv::Rodrigues(Rvec, rotMat);
		// Copy to transformation matrix
		for (int col = 0; col < 3; col++)
		{
			for (int row = 0; row < 3; row++)
			{
				m.rotMatrix(row,col) = rotMat(row, col); // Copy	rotation component
			}
			m.translation.data[col] = Tvec(col); // Copy translation	component
		}
		// Since solvePnP finds camera location, w.r.t to marker pose, to get marker pose w.r.t to the camera we invert it.
		m.translation = -m.translation;
		m.rotMatrix = m.rotMatrix.inv();		
	}
}